@@ -931,7 +931,47 @@ ogrGetForeignPaths(PlannerInfo* root,
 }
 
 
+/*
+ * Convert an OgrFdwSpatialFilter into a List so it can
+ * be safely passed through the fdw_private list.
+ */
+static List*
+ogrSpatialFilterToList(const OgrFdwSpatialFilter* spatial_filter)
+{
+	List *l = NIL;
+	if (spatial_filter)
+	{
+		l = lappend(l, makeInteger(spatial_filter->ogrfldnum));
+		l = lappend(l, makeFloat(psprintf("%.17g", spatial_filter->minx)));
+		l = lappend(l, makeFloat(psprintf("%.17g", spatial_filter->miny)));
+		l = lappend(l, makeFloat(psprintf("%.17g", spatial_filter->maxx)));
+		l = lappend(l, makeFloat(psprintf("%.17g", spatial_filter->maxy)));
+	}
+	return l;
+}
 
+/*
+ * Convert the List form back into an OgrFdwSpatialFilter
+ * after passing through fdw_private.
+ */
+static OgrFdwSpatialFilter*
+ogrSpatialFilterFromList(const List* lst)
+{
+	OgrFdwSpatialFilter* spatial_filter;
+
+	if (lst == NIL)
+		return NULL;
+
+	Assert(list_length(lst) == 5);
+
+	spatial_filter = palloc(sizeof(OgrFdwSpatialFilter));
+	spatial_filter->ogrfldnum = intVal(linitial(lst));
+	spatial_filter->minx = floatVal(lsecond(lst));
+	spatial_filter->miny = floatVal(lthird(lst));
+	spatial_filter->maxx = floatVal(lfourth(lst));
+	spatial_filter->maxy = floatVal(list_nth(lst, 4)); /* list_nth counts from zero */
+	return spatial_filter;
+}
 
 /*
  * fileGetForeignPlan
@@ -956,6 +996,8 @@ ogrGetForeignPlan(PlannerInfo* root,
 	List* fdw_private;
 	OgrFdwPlanState* planstate = (OgrFdwPlanState*)(baserel->fdw_private);
 	OgrFdwState* state = (OgrFdwState*)(baserel->fdw_private);
+	OgrFdwSpatialFilter* spatial_filter = NULL;
+	char* attribute_filter = NULL;
 
 	/* Add in column mapping data to build SQL with the right OGR column names */
 	ogrReadColumnData(state);
@@ -966,9 +1008,19 @@ ogrGetForeignPlan(PlannerInfo* root,
 	 * down to OGR via OGR_L_SetAttributeFilter (done) and (TODO) OGR_L_SetSpatialFilter.
 	 */
 	initStringInfo(&sql);
-	sql_generated = ogrDeparse(&sql, root, baserel, scan_clauses, state, &params_list);
-	elog(DEBUG1, "OGR SQL: %s", sql.data);
+	sql_generated = ogrDeparse(&sql, root, baserel, scan_clauses, state, &params_list, &spatial_filter);
 
+	/* Extract the OGR SQL from the StringInfoData */
+	if (sql_generated && sql.len > 0)
+		attribute_filter = sql.data;
+
+	/* Log filters at debug level one as necessary */
+	if (attribute_filter)
+		elog(DEBUG1, "OGR SQL: %s", attribute_filter);
+	if (spatial_filter)
+		elog(DEBUG1, "OGR spatial filter (%g %g, %g %g)",
+		             spatial_filter->minx, spatial_filter->miny,
+		             spatial_filter->maxx, spatial_filter->maxy);
 	/*
 	 * Here we strip RestrictInfo
 	 * nodes from the clauses and ignore pseudoconstants (which will be
@@ -982,25 +1034,13 @@ ogrGetForeignPlan(PlannerInfo* root,
 	 */
 	scan_clauses = extract_actual_clauses(scan_clauses, false);
 
-	/*
-	 * Serialize the data we want to pass to the execution stage.
-	 * This is ugly but seems to be the only way to pass our constructed
-	 * OGR SQL command to execution.
-	 *
-	 * TODO: Pass a spatial filter down also.
-	 */
-	if (sql_generated)
-	{
-		fdw_private = list_make2(makeString(sql.data), params_list);
-	}
-	else
-	{
-		fdw_private = list_make2(NULL, params_list);
-	}
+	/* Pack the data we want to pass to the execution stage into a List. */
+	/* The members of this list must by copyable by PgSQL, which means */
+	/* they need to be Lists themselves, or Value nodes, otherwise when */
+	/* the plan gets copied the copy might fail. */
+	fdw_private = list_make3(makeString(attribute_filter), params_list, ogrSpatialFilterToList(spatial_filter));
 
-	/*
-	 * Clean up our connection
-	 */
+	/* Clean up our connection */
 	ogrFinishConnection(&(planstate->ogr));
 
 	/* Create the ForeignScan node */
@@ -1207,7 +1247,7 @@ ogrBytesToHex(unsigned char* bytes, size_t size)
 	int i;
 	if (! bytes || ! size)
 	{
-		elog(ERROR, "hexbytes_from_bytes: invalid input");
+		elog(ERROR, "ogrBytesToHex: invalid input");
 		return NULL;
 	}
 	hex = palloc(size * 2 + 1);
@@ -1293,7 +1333,12 @@ ogrReadColumnData(OgrFdwState* state)
 	tbl = palloc0(sizeof(OgrFdwTable));
 
 	/* One column for each PgSQL foreign table column */
+#if PG_VERSION_NUM < 120000
 	rel = heap_open(state->foreigntableid, NoLock);
+#else
+	rel = table_open(state->foreigntableid, NoLock);
+#endif /* PG_VERSION_NUM */
+
 	tupdesc = rel->rd_att;
 	state->tupdesc = tupdesc;
 	tbl->ncols = tupdesc->natts;
@@ -1453,7 +1498,12 @@ ogrReadColumnData(OgrFdwState* state)
 		}
 	}
 	pfree(ogr_fields);
+#if PG_VERSION_NUM < 120000
 	heap_close(rel, NoLock);
+#else
+	table_close(rel, NoLock);
+#endif /* PG_VERSION_NUM */
+
 
 	return;
 }
@@ -1511,12 +1561,19 @@ ogrLookupGeometryFunctionOid(const char* proname)
 static void
 ogrBeginForeignScan(ForeignScanState* node, int eflags)
 {
+	OgrFdwState* state;
+	OgrFdwExecState* execstate;
+	OgrFdwSpatialFilter* spatial_filter;
 	Oid foreigntableid = RelationGetRelid(node->ss.ss_currentRelation);
 	ForeignScan* fsplan = (ForeignScan*)node->ss.ps.plan;
 
+	/* Do nothing in EXPLAIN (no ANALYZE) case */
+	if (eflags & EXEC_FLAG_EXPLAIN_ONLY)
+		return;
+
 	/* Initialize OGR connection */
-	OgrFdwState* state = getOgrFdwState(foreigntableid, OGR_EXEC_STATE);
-	OgrFdwExecState* execstate = (OgrFdwExecState*)state;
+	state = getOgrFdwState(foreigntableid, OGR_EXEC_STATE);
+	execstate = (OgrFdwExecState*)state;
 
 	/* Read the OGR layer definition and PgSQL foreign table definitions */
 	ogrReadColumnData(state);
@@ -1525,10 +1582,26 @@ ogrBeginForeignScan(ForeignScanState* node, int eflags)
 	execstate->setsridfunc = ogrLookupGeometryFunctionOid("st_setsrid");
 	execstate->typmodsridfunc = ogrLookupGeometryFunctionOid("postgis_typmod_srid");
 
-	/* Get private info created by planner functions. */
-	execstate->sql = strVal(list_nth(fsplan->fdw_private, 0));
+	/* Get OGR SQL generated by the deparse step during the planner function. */
+	execstate->sql = (char*) strVal(list_nth(fsplan->fdw_private, 0));
+
+	/* TODO: Use the parse step attribute list to restrict requested columns */
 	// execstate->retrieved_attrs = (List *) list_nth(fsplan->fdw_private, 1);
 
+	/* Get spatial filter generated by the deparse step. */
+	spatial_filter = ogrSpatialFilterFromList(list_nth(fsplan->fdw_private, 2));
+
+	if (spatial_filter)
+	{
+		OGR_L_SetSpatialFilterRectEx(execstate->ogr.lyr,
+		                             spatial_filter->ogrfldnum,
+		                             spatial_filter->minx,
+		                             spatial_filter->miny,
+		                             spatial_filter->maxx,
+		                             spatial_filter->maxy
+		                             );
+	}
+
 	if (execstate->sql && strlen(execstate->sql) > 0)
 	{
 		OGRErr err = OGR_L_SetAttributeFilter(execstate->ogr.lyr, execstate->sql);
@@ -1944,6 +2017,20 @@ ogrEwkbStripSrid(unsigned char* wkb, size_t wkbsize)
 	return newwkbsize;
 }
 
+OGRErr
+pgDatumToOgrGeometry (Datum pg_geometry, Oid pgsendfunc, OGRGeometryH* ogr_geometry)
+{
+	OGRErr err;
+	bytea* wkb_bytea = DatumGetByteaP(OidFunctionCall1(pgsendfunc, pg_geometry));
+	unsigned char* wkb = (unsigned char*)VARDATA_ANY(wkb_bytea);
+	size_t wkbsize = VARSIZE_ANY_EXHDR(wkb_bytea);
+	wkbsize = ogrEwkbStripSrid(wkb, wkbsize);
+	err = OGR_G_CreateFromWkb(wkb, NULL, ogr_geometry, wkbsize);
+	if (wkb_bytea)
+		pfree(wkb_bytea);
+	return err;
+}
+
 static OGRErr
 ogrSlotToFeature(const TupleTableSlot* slot, OGRFeatureH feat, const OgrFdwTable* tbl)
 {
@@ -2037,21 +2124,9 @@ ogrSlotToFeature(const TupleTableSlot* slot, OGRFeatureH feat, const OgrFdwTable
 			else
 			{
 				OGRGeometryH geom;
-				bytea* wkb_bytea = DatumGetByteaP(OidFunctionCall1(col.pgsendfunc, values[i]));
-				unsigned char* wkb = (unsigned char*)VARDATA_ANY(wkb_bytea);
-				int wkbsize = VARSIZE_ANY_EXHDR(wkb_bytea);
-				wkbsize = ogrEwkbStripSrid(wkb, wkbsize);
-
-				/* TODO, create geometry with SRS of table? */
-				err = OGR_G_CreateFromWkb(wkb, NULL, &geom, wkbsize);
-				if (wkb_bytea)
-				{
-					pfree(wkb_bytea);
-				}
+				err = pgDatumToOgrGeometry (values[i], col.pgsendfunc, &geom);
 				if (err != OGRERR_NONE)
-				{
 					return err;
-				}
 
 #if (GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(1,11,0))
 				err = OGR_F_SetGeomFieldDirectly(feat, ogrfldnum, geom);
@@ -2305,10 +2380,11 @@ static void
 ogrEndForeignScan(ForeignScanState* node)
 {
 	OgrFdwExecState* execstate = (OgrFdwExecState*) node->fdw_state;
-
-	elog(DEBUG2, "processed %d rows from OGR", execstate->rownum);
-
-	ogrFinishConnection(&(execstate->ogr));
+	if (execstate)
+	{
+		elog(DEBUG2, "OGR FDW processed %d rows from OGR", execstate->rownum);
+		ogrFinishConnection(&(execstate->ogr));
+	}
 
 	return;
 }
